#include "music.h"
#include "main.h"
#include "playlist.h"

#include <gst/gconf/gconf.h>
#include <gst/gst.h>
#include <glib-2.0/glib.h>
#include <string.h>
#include <stdarg.h>

static GstElement *pipeline, *source, *decode, *sink;
static GstElement *inqueue, *outqueue, *inthread, *outthread;
static GList *notify_channels = NULL;

static gboolean remote_mode = FALSE;
static gboolean using_gnomevfs = FALSE;

gboolean music_playing = FALSE;

gboolean
music_flush (gpointer data)
{
    gst_bin_iterate (GST_BIN(data));
    return music_playing;
}

void 
music_eos (GstElement *element, gpointer data) 
{
    playlist_advance (1, main_loop_at_end);
}

/* This is the layout of our pipeline:
   for remote:
   { { src ! queue } ! spider ! { queue ! sink } }
   for local:
   { src ! spider ! { queue ! sink } }
 
   I copied it from the GstPlay library's hyperthreaded model, and added
   a queue at the front to allow streaming from latent sources.
*/
void
music_init ()
{
    pipeline = gst_pipeline_new ("pipe");
    g_assert (pipeline);

    inthread = gst_thread_new ("inthread");
    g_assert (inthread);
    gst_object_ref (GST_OBJECT (inthread));

    inqueue = gst_element_factory_make ("queue", "inqueue");
    g_assert (inqueue);

    if (!(source = gst_element_factory_make ("gnomevfssrc", "source"))) {
        g_warning ("unable to create gnome-vfs stream, using standard "
                   "file stream");
        if (!(source = gst_element_factory_make ("filesrc", "source"))) {
            g_critical ("unable to create source");
            exit (EXIT_FAILURE);
        }
    } else {
        using_gnomevfs = TRUE;
    }
    gst_object_ref (GST_OBJECT (source));

    outthread = gst_thread_new ("outthread");
    g_assert (outthread);

    outqueue = gst_element_factory_make ("queue", "outqueue");
    g_assert (outqueue);

    if (!(sink = gst_gconf_get_default_audio_sink ())) {
        if (!(sink = gst_element_factory_make ("alsasink", "sink"))) {
            if (!(sink = gst_element_factory_make ("osssink", "sink"))) {
                g_critical ("unable to create output plugin");
                exit (EXIT_FAILURE);
            }
        }
    }
    g_signal_connect (G_OBJECT (sink), "eos", G_CALLBACK (music_eos),
                      pipeline);

    gst_bin_add (GST_BIN (inthread), inqueue);

    gst_bin_add (GST_BIN (pipeline), outthread);
    gst_bin_add (GST_BIN (outthread), outqueue);
    gst_bin_add (GST_BIN (outthread), sink);

    gst_element_link (outqueue, sink);

    decode = NULL;
}


void
music_destroy ()
{ 
    gst_object_unref (GST_OBJECT (source));
    gst_object_unref (GST_OBJECT (inthread));

    gst_element_set_state (GST_ELEMENT (pipeline), GST_STATE_NULL);
    gst_object_unref (GST_OBJECT (pipeline));
}

static void
pipeline_setup (gboolean remote)
{
    gboolean setup = (decode != NULL);

    g_message ("%d %d %p %d", remote, remote_mode, decode, using_gnomevfs);

    /* have we set up the pipeline yet? */
    if (setup) {
        /* replace the spider */
        if (remote_mode) {
            gst_element_unlink (decode, inqueue);

            if (!remote) {
                gst_element_unlink (source, inqueue);

                gst_bin_remove (GST_BIN (inthread), source);
                gst_bin_remove (GST_BIN (pipeline), inthread);
            }
        } else {
            gst_element_unlink (decode, source);

            if (remote)
                gst_bin_remove (GST_BIN (pipeline), source);
        }

        gst_element_unlink (decode, outqueue);
        gst_bin_remove (GST_BIN (pipeline), decode);
    }
    /* XXX this is a fucking disgusting hack to work around the bug that is
       known as the "spider" plugin. It seems that it is the best form of
       auto-detection are not as good, but ugh. so here we are, all streams
       are decoded with "mad" because vorbis won't work with our remote
       pipeline layout anyways (more bugs!). when gst fixes these, i will be
       happy to let you play vorbis/etc streams!

       XXX note: when the vorbis bug is fixed, this complicated pipeline
       shifting shit should be reverted to just using the remote pipeline
       all the time. its revision #183 for music.c in the svn repository.
    */
    if (!(decode = gst_element_factory_make ((remote ? "mad" : "spider"),
                                             "decoder"))) {
        music_notify ("unable to instantiate the spider plugin");
        exit (EXIT_FAILURE);
    }
    gst_bin_add (GST_BIN (pipeline), decode);
    gst_element_link (decode, outqueue);

    if (using_gnomevfs)
        g_object_set (G_OBJECT (source), "iradio-mode", remote, NULL);

    if (remote) {
        if (!remote_mode || !setup) {
            gst_bin_add (GST_BIN (inthread), source);
            gst_bin_add (GST_BIN (pipeline), inthread);

            gst_element_link (source, inqueue);
        }

        gst_element_link (inqueue, decode);
    } else {
        if (remote_mode || !setup) {
            gst_bin_add (GST_BIN (pipeline), source);
        }

        gst_element_link (source, decode);
    }

    remote_mode = remote;
}

void
music_play ()
{
    struct PlaylistItem *item;
    GstElementState state;
    gboolean play;

    state = gst_element_get_state(GST_ELEMENT(pipeline));

    play = state != GST_STATE_PLAYING && playlist_current;

    if (play && state != GST_STATE_PAUSED) {
        item = LISTITEM (playlist_current);

        music_notify ("playing: %s", item->utf8_path);

        if (state != GST_STATE_READY)
            gst_element_set_state (GST_ELEMENT (pipeline), GST_STATE_READY);

        pipeline_setup (item->type == TYPE_URL);

        g_object_set (G_OBJECT (source), "location", item->path, NULL);

        g_object_set (G_OBJECT (inqueue), "max_level",
                      main_inqueue_buffers, NULL);
        g_object_set (G_OBJECT (outqueue), "max_level",
                      main_outqueue_buffers, NULL);
    }
    if (play) {
        if (gst_element_set_state (GST_ELEMENT (pipeline),
                                   GST_STATE_PLAYING) != GST_STATE_FAILURE) {
            music_playing = TRUE;
            g_idle_add (music_flush, pipeline);
        } else {
            /* remove it from the playlist */
            playlist_remove (g_list_position (playlist, playlist_current));
        }
    }
}

void
music_pause ()
{
    GstElementState state = gst_element_get_state (GST_ELEMENT (pipeline));
    if (state == GST_STATE_PLAYING)
        gst_element_set_state (GST_ELEMENT (pipeline), GST_STATE_PAUSED);
    music_playing = FALSE;
}

void
music_stop ()
{
    gst_element_set_state (GST_ELEMENT (pipeline), GST_STATE_READY);
    music_playing = FALSE;
}

void
music_add_notify (GIOChannel *output)
{
    notify_channels = g_list_prepend (notify_channels, output);
}

gboolean
music_list_notify (GIOChannel *channel, const char *message)
{
    gsize written;
    GIOStatus status;
    gboolean retval = FALSE;

    while ( (status = g_io_channel_write_chars (channel, message,
                                                strlen (message),
                                                &written, NULL))
            == G_IO_STATUS_AGAIN) {
        message += written;
    }
    
    if (status == G_IO_STATUS_ERROR) {
        retval = TRUE;
    }
    while ( (status = g_io_channel_flush(channel, NULL)) == G_IO_STATUS_AGAIN);

    if (status == G_IO_STATUS_ERROR)
        retval = TRUE;

    return retval;
}
    
void
music_notify (const char *format, ...)
{
    char *message, *newformat;
    GList *it;
    va_list args;

    va_start (args, format);
    
    message = g_strdup_vprintf(format, args);
    g_message ("%s", message);
    g_free (message);
            
    newformat = g_strdup_printf ("%s\n", format);
    message = g_strdup_vprintf (newformat, args);
    it = notify_channels;

    while (it != NULL) {
        if (music_list_notify (it->data, message)) {
            GList *tmp = it->next;
            g_list_remove_link (notify_channels, it);
            g_io_channel_shutdown (it->data, FALSE, NULL);
            g_io_channel_unref (it->data);
            if (it == notify_channels) {
                notify_channels = NULL;
            }
            it = tmp;
        } else {
            it = it->next;
        }
    }
    g_free (message);
    g_free (newformat);
    
    va_end (args);
}
