#include "playlist.h"
#include "music.h"
#include "main.h"

#include <string.h>
#include <sys/types.h> /* for stat() */
#include <sys/stat.h>  /* for stat() */
#include <unistd.h>    /* for stat() */
#include <errno.h>

GList *playlist = NULL;
GList *playlist_current = NULL;

static GList *playlist_random = NULL;

void
playlist_init ()
{
}

void
playlist_destroy ()
{
    playlist_clear ();
}

struct PlaylistItem *listitem_new(char *f, char *u, PlaylistItemType type)
{
    struct PlaylistItem *s;

    s = g_new (struct PlaylistItem, 1);
    s->path = f;
    s->utf8_path = u;
    s->type = type;
    return s;
}

void listitem_free (struct PlaylistItem *item)
{
    g_free (item->utf8_path);
    g_free (item->path);
    g_free (item);
}

void
playlist_append (const char *path)
{
    gchar *f, *u;
    struct PlaylistItem *item;
    PlaylistItemType type;
    gint p;
    GError *e = NULL;
    GDir *dir;
    struct stat st;

    /* XXX errors encountered here should go thru the client fifos */

    if (g_utf8_validate (path, -1, NULL)) {
        if (!(f = g_filename_from_utf8 (path, -1, NULL, NULL, NULL))) {
            g_critical ("Skipping '%s'. Could not convert from UTF-8. Bug?",
                        path);
            return;
        }
        u = g_strdup (path);
    } else {
        if (!(u = g_filename_to_utf8 (path, -1, NULL, NULL, NULL))) {
            g_warning ("Skipping '%s'. Could not convert to UTF-8. "
                       "See the README for a possible solution.", path);
            return;
        }
        f = g_strdup (path);
    }

    /* XXX use stat here lots ktnx */

    /* am i a url? */
    if (g_str_has_prefix (u, "http://")) {
        /* yay i'm a url... */
        type = TYPE_URL;
    } else {
        type = TYPE_FILE;

        /* figure out if i am a directory and recurse */
        if (stat (f, &st) < 0) {
            g_warning ("Skipping '%s'. Could not stat the location. The error "
                       "given was '%s'.", u, strerror(errno));
            g_free (f);
            g_free (u);
            return;
        }
        if (S_ISDIR (st.st_mode)) {
            /* first, kill the trailing / for consistancy */
            if (g_str_has_suffix (u, "/")) {
                f[strlen (f) - 1] = '\0';
                u[strlen (u) - 1] = '\0';
            }

            dir = g_dir_open (u, 0, &e);
            if (!dir) {
                g_warning ("Unable to open the directory '%s'. The error "
                           "given was: %s", u, e->message);
                g_clear_error (&e);
            } else {
                GSList *list = NULL, *it;
                const gchar *n;

                while ((n = g_dir_read_name (dir)))
                    list = g_slist_prepend (list,
                                            g_strconcat (u, "/", n, NULL));
                list = g_slist_reverse (list);
                g_dir_close (dir);

                /* alphabatize these pls */
                list = g_slist_sort (list, (GCompareFunc) strcmp);
                for (it = list; it; it = g_slist_next (it)) {
                    playlist_append (it->data);
                    g_free (it->data);
                }
                g_slist_free (list);
            }

            g_free (f);
            g_free (u);
            return;
        } else {
            /* am i a playlist? */
            if (g_str_has_suffix (u, ".m3u") || g_str_has_suffix (u, ".pls")) {
                GIOChannel *ch = g_io_channel_new_file (u, "r", &e);
                if (e) {
                    g_warning ("Skipping '%s'. Unable to open the playlist. "
                               "The error given was: %s", u, e->message);
                    g_error_free (e);
                    g_free (f);
                    g_free (u);
                    return;
                } else {
                    gchar *song;
                    gint term;
                    while (g_io_channel_read_line (ch, &song, NULL, &term, &e)
                           == G_IO_STATUS_NORMAL) {
                        song[term] = '\0';
                        if (!g_path_is_absolute (song)) {
                            gchar *dir;
                            gchar *full;

                            dir = g_path_get_dirname (u);
                            full = g_strconcat (dir, "/", song, NULL);
                            g_free (dir);
                            g_free (song);
                            song = full;
                        }
                        playlist_append (song);
                        g_free (song);
                    }
                    if (e) {
                        g_warning ("Skipping the remainder of the playlist "
                                   "'%s'. The error given was: %s",
                                   u, e->message);
                        g_error_free (e);
                    }
                    g_io_channel_unref (ch);
                    g_free (f);
                    g_free (u);
                    return;
                }
            }
        }
    }

    item = listitem_new(f, u, type);
    music_notify ("appending %s", item->utf8_path);
    playlist = g_list_append (playlist, item);

    if (!playlist_current) playlist_current = playlist;

    p = g_list_position (playlist_random,
                         g_list_find (playlist_random,
                                      playlist_current->data));
    playlist_random = g_list_insert
        (playlist_random, item, 
         g_random_int_range (p, g_list_length (playlist)) + 1);
}

void
playlist_advance (int num, gboolean loop)
{
    static GList *rand_cur = NULL;
    gboolean looped = FALSE;

    music_notify ("advancing %d track(s)", num);

    if (!playlist) return;
    
    if (main_random_order) {
        rand_cur = g_list_find (playlist_random, playlist_current->data);

        while (rand_cur && num > 0) {
            rand_cur = g_list_next (rand_cur);
            if (!rand_cur) {
                /* XXX re-randomize the whole thing */
                rand_cur = playlist_random;
                looped = TRUE;
            }
            --num;
        }
        while (rand_cur && num < 0) {
            rand_cur = g_list_previous (rand_cur);
            if (!rand_cur) {
                /* XXX re-randomize the whole thing */
                rand_cur = g_list_last (playlist_random);
                looped = TRUE;
            }
            ++num;
        }
        playlist_current = g_list_find (playlist, rand_cur->data);
    } else {
        while (playlist_current && num > 0) {
            playlist_current = g_list_next (playlist_current);
            if (!playlist_current) {
                playlist_current = playlist;
                looped = TRUE;
            }
            --num;
        }
        while (playlist_current && num < 0) {
            playlist_current = g_list_previous (playlist_current);
            if (!playlist_current) {
                playlist_current = g_list_last (playlist);
                looped = TRUE;
            }
            ++num;
        }
    }
    music_stop ();
    if (!looped || loop)
        music_play ();
}

void
playlist_seek (int num)
{
    GList *it = g_list_nth (playlist, num);

    if (it) {
        music_notify ("seeking to track %d", num);
        
        playlist_current = it;
        /* this function is used during load, and we don't want to start
           playing necessarily */
        if (main_status == QUARK_RUNNING) {
            music_stop ();
            music_play ();
        }
    }
}

void
playlist_clear ()
{
    music_notify ("clearing playlist");
    
    while (playlist) {
        listitem_free (playlist->data);
        playlist = g_list_delete_link (playlist, playlist);
    }
    playlist = playlist_current = NULL;
    g_list_free (playlist_random);
    playlist_random = NULL;
    music_stop ();
}

void
playlist_remove (int num)
{
    GList *it = g_list_nth (playlist, num);

    if (it) {
        music_notify ("removing track %d", num);
        
        if (it == playlist_current)
            playlist_advance (1, main_loop_at_end);
        if (it == playlist_current) {
            music_stop ();
            playlist_current = NULL;
        }

        listitem_free (it->data);
        playlist = g_list_delete_link (playlist, it);
    }
}

void
playlist_move (int num, int before)
{
    GList *it = g_list_nth (playlist, num);
    struct PlaylistItem *item;

    if (!it || num == before || num == before + 1) return;

    music_notify ("moving track %d before track %d", num, before);
    
    item = it->data;
    playlist = g_list_insert_before (playlist, g_list_nth (playlist, before),
                                     item);
    playlist = g_list_delete_link(playlist, it);
    /* XXX reinsert in the random order list? */
}

void
playlist_dump ()
{
    GList *it;
    int i = 1;

    music_notify ("Current playlist:");

    for (it = playlist; it; it = g_list_next(it)) {
        char *base = g_path_get_basename
            (((struct PlaylistItem*)it->data)->utf8_path);
        g_print ("%s %d %s\n",
                 (it->data == playlist_current->data ? ">" : " "), i++, base);
        g_free (base);
    }
}
